#include <cstdio>
#include <cstdlib>
#include <iostream>
#include "pcl/io/pcd_io.h"
#include "pcl/point_types.h"

#include "pcl/features/normal_3d.h"
#include "pcl/features/fpfh.h"
#include "pcl/registration/registration.h"
#include "pcl/registration/icp.h"
#include "pcl/registration/icp_nl.h"
#include "pcl/registration/ia_ransac.h"

//for statrm -- this conflicts with flann in openCV < 2.2; just upgrade opencv
#include "pcl/filters/statistical_outlier_removal.h"
#include "pcl/filters/voxel_grid.h"

#include <cmath>

typedef struct
{
  int scanNum;
  char filename[50];
  double posX;
  double posY;
  double head;
  int pan;
  int tilt;
} logRead;

int bufferToRead(char* buffer,logRead* myRead)
{
  //0 robot0-10839334-S0P-20.txt 0.000000, 0.000000, 0.000000,  -20, 0
  int count = sscanf(buffer,"%d %s %lf, %lf, %lf, %d, %d",
		     &myRead->scanNum,myRead->filename,&myRead->posX,&myRead->posY,&myRead->head,&myRead->pan,&myRead->tilt);
  if(count ==7)
    return 1;    //success
  else
    {
      printf("count in b2r: %d",count);
      return 0;    //failure
    }
}

void printLogRead(logRead* myRead)
{
  printf("%d %s %lf, %lf, %lf, %d, %d",
		     myRead->scanNum,myRead->filename,myRead->posX,myRead->posY,myRead->head,myRead->pan,myRead->tilt);
}

void transform2pc(pcl::PointCloud<pcl::PointXYZ>& src, pcl::PointCloud<pcl::PointXYZ>& dst, logRead* cloudInfo)
{
  printf("parsing cloud info: \n");
  printLogRead(cloudInfo);

 //first rotate the cloud by the correct pan/tilt
  Eigen::Quaternionf tilt;
  //this might need a negative sign
  tilt = Eigen::AngleAxisf(M_PI/180.0*(cloudInfo->tilt),Eigen::Vector3f::UnitX());
  pcl::transformPointCloud(src,dst,Eigen::Vector3f(0.0,0.0,0.0),tilt);
  
  Eigen::Quaternionf pan;
  pan = Eigen::AngleAxisf(-M_PI/180.0*(cloudInfo->pan),Eigen::Vector3f::UnitY());
  pcl::transformPointCloud(dst,dst,Eigen::Vector3f(0.0,0.0,0.0),pan);

  //now transform to the robot coordinate frame
  Eigen::Matrix4f m;
  m.setIdentity();
  m << 0, 0, 1, 0, 
      -1, 0, 0, 0, 
       0,-1, 0, 0, 
       0, 0, 0, 1;
  
  pcl::transformPointCloud(dst,dst,m);

  Eigen::Quaternionf globalHeading;
  globalHeading = Eigen::AngleAxisf(M_PI/180.0*cloudInfo->head,Eigen::Vector3f::UnitZ());
  pcl::transformPointCloud(dst,dst,Eigen::Vector3f(cloudInfo->posX/1000.0,cloudInfo->posY/1000.0,0.0),globalHeading);

  printf("Done transforming point cloud");

}


int main(int argc, char** argv)
{
  if(argc==0)
    {printf("warning, not enough command line arguments."); exit(0);}

  //read in list of transformations and filenames
  FILE* log = fopen(argv[1],"r");
  char buffer[100];

  logRead reading;

  pcl::PointCloud<pcl::PointXYZ> bigCloud,currentCloudSrc,currentCloudDst;
  pcl::PCDReader reader;

  fgets(buffer,sizeof(buffer),log);
  printf("Reading log file: %s",buffer);

  int count = 0;
  while( fgets(buffer,sizeof(buffer),log ) != NULL)
    {
      printf("%s\n",buffer);

      if(!bufferToRead(buffer,&reading))
	{
	  printf("no buffer to read");
	  break;
	}
      else
	{
	  count++;
	  strcat(reading.filename,".pcd");

	  printf("Opening file %s\n",reading.filename);
	  reader.read<pcl::PointXYZ>(reading.filename,currentCloudSrc);
	  std::cout << currentCloudSrc << std::endl;
	  transform2pc(currentCloudSrc,currentCloudSrc,&reading);
	  
	  if(count==1)
	    bigCloud = currentCloudSrc;
	  else
	    {
	      bigCloud+=currentCloudSrc;
	    }
	}
    }
  fclose(log);

  pcl::PCDWriter writer;
  writer.write<pcl::PointXYZ>("hallway.pcd",bigCloud,false);

  return 0;
  //read in point cloud file names
  //read in odometry registrations
}
